#region using
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Design;
using System.Text;
using System.Xml;
#endregion
namespace Microsoft.ServiceModel.Samples.Transport
{
   public class TcpChannelListener
       : ChannelManagerBase, IChannelListener<IDuplexSessionChannel>
   {
      public TcpChannelListener(IMessageEncodingBindingElement encodingElement)
         : base()
      {
         if (encodingElement == null)
         {
            throw new ArgumentNullException("encodingElement");
         }

         this.encoderFactory = encodingElement.CreateMessageEncoderFactory();
         this.bufferManager =
             BufferManager.CreateBufferManager(512 * 1024, maxBufferSize);
      }

      public Type ChannelType
      {
         get { return typeof(IDuplexSessionChannel); }
      }

      #region Listening methods
      protected override void OnOpen(TimeSpan timeout)
      {
         Listen();
         base.OnOpen(timeout);
      }

      protected override IAsyncResult OnBeginOpen(TimeSpan timeout,
          AsyncCallback callback, object state)
      {
         Listen();
         return base.OnBeginOpen(timeout, callback, state);
      }

      void Listen()
      {
         if (uri == null)
         {
            throw new InvalidOperationException(
                "SetUri must be called before opening this Channel Listener");
         }

         IPEndPoint localEndpoint = new IPEndPoint(IPAddress.Any, uri.Port);

         this.listenSocket = new Socket(
             localEndpoint.AddressFamily, SocketType.Stream, ProtocolType.Tcp);

         this.listenSocket.Bind(localEndpoint);
         this.listenSocket.Listen(10);
      }
      #endregion

      #region Addressing
      public override string Scheme
      {
         get { return "my.tcp"; }
      }

      public Uri Uri
      {
         get { return this.uri; }
      }

      public void SetUri(Uri baseAddress, string relativeAddress)
      {
         #region parameter validation
         if (baseAddress == null)
            throw new ArgumentNullException("baseAddress");

         if (relativeAddress == null)
            throw new ArgumentNullException("relativeAddress");

         if (!baseAddress.IsAbsoluteUri)
         {
            throw new ArgumentException("Base address cannot be a relative Uri", "baseAddress");
         }

         if (baseAddress.Scheme != this.Scheme)
         {
            // URI schemes are case-insensitive, so try a case insensitive compare now
            if (string.Compare(baseAddress.Scheme, this.Scheme, StringComparison.OrdinalIgnoreCase) != 0)
            {
               throw new ArgumentException("Scheme for base address must be " + this.Scheme, "baseAddress");
            }
         }
         #endregion

         ThrowIfDisposedOrImmutable();
         this.uri = new Uri(baseAddress, relativeAddress);
      }

      public void SetUri(Uri uri)
      {
         SetUri(uri, "");
      }
      #endregion

      #region Accepting Server Channel
      public IDuplexSessionChannel AcceptChannel(TimeSpan timeout)
      {
         base.ThrowIfDisposedOrNotOpen();
         Socket dataSocket = null;
         lock (acceptLock)
         {
            dataSocket = listenSocket.Accept();
         }
         return new ServerTcpDuplexSessionChannel(this, dataSocket,
             new EndpointAddress(uri));
      }

      class ServerTcpDuplexSessionChannel : TcpDuplexSessionChannel
      {
         public ServerTcpDuplexSessionChannel(TcpChannelListener parent,
             Socket socket, EndpointAddress localAddress)
            : base(parent.encoderFactory, parent.bufferManager,
             TcpDuplexSessionChannel.AnonymousAddress,
             localAddress, TcpDuplexSessionChannel.AnonymousAddress.Uri, parent)
         {
            base.InitializeSocket(socket);
         }

         protected override void OnOpen(TimeSpan timeout)
         {
            VerifyPreamble();
            base.OnOpen(timeout);
         }

         protected override IAsyncResult OnBeginOpen(TimeSpan timeout,
             AsyncCallback callback, object state)
         {
            VerifyPreamble(); // would be asynchronous in production
            return base.OnBeginOpen(timeout, callback, state);
         }

         void VerifyPreamble()
         {
            Uri via = new Uri(base.ReadString());
            if (!LocalAddress.Uri.IsBaseOf(via))
            {
               throw new EndpointNotFoundException(
                   string.Format("Via {0} is not a prefix of local address {1}.",
                   via, LocalAddress.Uri));
            }

            string contentType = base.ReadString();
            if (!base.MessageEncoder.IsContentTypeSupported(contentType))
            {
               throw new CommunicationException(
                   string.Format("Content Type {0} was sent to a listener expecting {1}.",
                   contentType, MessageEncoder.ContentType));
            }
         }
      }
      #endregion

      #region IChannelListener Advanced Methods
      public override MessageVersion MessageVersion
      {
         get { return encoderFactory.MessageVersion; }
      }

      public Identity Identity
      {
         get { return null; }
      }

      public IDuplexSessionChannel AcceptChannel()
      {
         return this.AcceptChannel(DefaultReceiveTimeout);
      }

      public IAsyncResult BeginAcceptChannel(AsyncCallback callback, object state)
      {
         return this.BeginAcceptChannel(DefaultReceiveTimeout, callback, state);
      }

      public IAsyncResult BeginAcceptChannel(TimeSpan timeout, AsyncCallback callback, object state)
      {
         throw new Exception("The method or operation is not implemented.");
      }

      public IDuplexSessionChannel EndAcceptChannel(IAsyncResult result)
      {
         throw new Exception("The method or operation is not implemented.");
      }

      public bool WaitForChannel(TimeSpan timeout)
      {
         throw new Exception("The method or operation is not implemented.");
      }

      public IAsyncResult BeginWaitForChannel(TimeSpan timeout, AsyncCallback callback, object state)
      {
         throw new Exception("The method or operation is not implemented.");
      }

      public bool EndWaitForChannel(IAsyncResult result)
      {
         throw new Exception("The method or operation is not implemented.");
      }

      public void SetUniqueUri()
      {
         throw new Exception("The method or operation is not implemented.");
      }

      public void SetUniqueUri(string hostName)
      {
         throw new Exception("The method or operation is not implemented.");
      }

      public void SetUniqueUri(string hostName, int port)
      {
         throw new Exception("The method or operation is not implemented.");
      }

      protected override TimeSpan DefaultReceiveTimeout
      {
         get { return TimeSpan.FromMinutes(1); }
      }

      protected override TimeSpan DefaultSendTimeout
      {
         get { return TimeSpan.FromMinutes(1); }
      }

      protected override TimeSpan DefaultCloseTimeout
      {
         get { return TimeSpan.FromMinutes(1); }
      }

      protected override TimeSpan DefaultOpenTimeout
      {
         get { return TimeSpan.FromMinutes(1); }
      }
      #endregion

      #region member variables
      const int maxBufferSize = 64 * 1024;
      BufferManager bufferManager;
      MessageEncoderFactory encoderFactory;
      Socket listenSocket;
      object acceptLock = new object();
      Uri uri;
      #endregion
   }
}